#------------------------------------------------------------------
# $Copyright (c) 2018 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other
# names may be trademarks of their respective owners.
# 
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/InfoSphere/InfoSphere.pm,v 1.12 2012/11/30 11:08:42 kkorde Exp $
#
# $sAuthor: $
#
# $Date: 2012/11/30 11:08:42 $
# $Revision: 1.12 $
#
# Provides:                            InfoSphere
# Author:                              kkorde
# crAgent Build:                       \$Id: crAgent,v 1.100 201
# crAgent Run Date:                    26 Jul, 2012
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package InfoSphere;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::Basename;
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: InfoSphere.pm,v 1.12 2012/11/30 11:08:42 kkorde Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define InfoSphere subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String Component,
                              IN String User,
                              IN String InfoSphereInstanceDir,
                              IN String EnvFile,
                              IN String InfoSphereInstance,
                              IN String Hostname,
                              IN Integer Port,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localComponent                The engine tier component that we want to start.
 $localUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 $localInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $localEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $localInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $localHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $localPort                     Port allocated for the specific engine tier component. 
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localComponent,
       $localUser,
       $localInfoSphereInstanceDir,
       $localEnvFile,
       $localInfoSphereInstance,
       $localHostname,
       $localPort,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  Component                [%s]
  User                     [%s]
  InfoSphereInstanceDir    [%s]
  EnvFile                  [%s]
  InfoSphereInstance       [%s]
  Hostname                 [%s]
  Port                     [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,    $sEntryPointName,    $localComponent,    $localUser,
  $localInfoSphereInstanceDir,                       $localEnvFile,
  $localInfoSphereInstance,       $localHostname,       $localPort,
  $localSecondLevelMonitor, $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  Component  User   InfoSphereInstanceDir   EnvFile
       InfoSphereInstance     Hostname      Port      SecondLevelMonitor
       MonitorProgram );

  #---------------------------------
  # list of supported components ...
  #---------------------------------
  my @lsSupportedComponents = qw ( DSEngine ASBNodeAgent ASBLoggingAgent );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs = ();
  my @lsOfflineArgs = ();
  my @lsMonitorArgs = ();
  my @lsCleanArgs = ();
  my @lsImfRegisterArgs = ();

  #--------------------------------
  # Check for invalid component ...
  #--------------------------------
  if ( ! grep ( /\b$localComponent\b/, @lsSupportedComponents ) ) {
    &logError ( 20168, "<%s> Component [%s] is an invalid or unsupported component", $sMyName, $localComponent);
    $iReturnCode = 0; # Failure..
  }

  if($localComponent eq "DSEngine"){

     @lsOnlineArgs  = qw ( InfoSphereInstanceDir Component User EnvFile );
     @lsOfflineArgs = qw ( InfoSphereInstanceDir Component User EnvFile );
     @lsMonitorArgs = qw ( InfoSphereInstanceDir Component User InfoSphereInstance );
     @lsCleanArgs   = qw ( InfoSphereInstanceDir Component User InfoSphereInstance );
     @lsImfRegisterArgs = qw ( InfoSphereInstanceDir Component User Port InfoSphereInstance );

  }else{

     @lsOnlineArgs  = qw ( InfoSphereInstanceDir Component User );
     @lsOfflineArgs = qw ( InfoSphereInstanceDir Component User );
     @lsMonitorArgs = qw ( InfoSphereInstanceDir Component User );
     @lsCleanArgs   = qw ( InfoSphereInstanceDir Component User );
     @lsImfRegisterArgs = qw ( InfoSphereInstanceDir Component User Port );

  }

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs,
	            'imf_register'  => \@lsImfRegisterArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  if ( $sEntryPointName eq "online"  ||
       $sEntryPointName eq "offline" ){

       #-----------------------------------------
       # Validate parameter InfoSphereInstanceDir..
       #-----------------------------------------
       if ( ! -d $localInfoSphereInstanceDir ) {
           &logError ( 20003,
           "<%s> InfoSphereInstanceDir [%s] directory does not exist",
           $sMyName, $localInfoSphereInstanceDir );
           $iReturnCode = 0; # Failure..
       }

       #---------------------------
       # Validate parameter EnvFile..
       #---------------------------
       if ( $localComponent eq "DSEngine" ){
            if ( ! -f $localEnvFile ) {
                 &logError ( 20004, "<%s> EnvFile [%s] is not a plain file", $sMyName, $localEnvFile );
                 $iReturnCode = 0; # Failure..
            }
       }
  }

  #------------------------
  # Validate parameter User..
  #------------------------
  if ( ! defined getpwnam $localUser ) {
     &logError ( 20005,
      "<%s> User [%s] is not a valid UNIX login",
     $sMyName, $localUser );
     $iReturnCode = 0; # Failure..
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20006,
      "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

  }

  $iReturnCode ||
  &logWarn ( 20007,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String Component,
                         IN String EnvFile,
                         IN String Hostname,
                         IN String InfoSphereInstance,
                         IN String InfoSphereInstanceDir,
                         IN Integer Port,
                         IN Integer SecondLevelMonitor,
                         IN String User );

 Description : Start logic needed for online entry pt

 Parameters :

 $sComponent                The engine tier component that we want to start.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sEnvFile,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $iPort,
       $iSecondLevelMonitor,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20008, "Subroutine <%s> called with:
  Component                [%s]
  EnvFile                  [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  SecondLevelMonitor       [%s]
  User                     [%s]",

  $sMyName,       $sComponent,        $sEnvFile,        $sHostname,
  $sInfoSphereInstance,      $sInfoSphereInstanceDir,       $iPort,
  $iSecondLevelMonitor, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  
  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting InfoSphere processes..
  #-----------------------------------------
  &logTrace ( 20009,
  "<%s> Killing InfoSphere processes for this instance", $sMyName );

  if ( InfoSphere::KillInfoSphereProcesses ( $sComponent,
                                             $sHostname,
                                             $sInfoSphereInstance,
                                             $sInfoSphereInstanceDir,
                                             $iPort,
                                             $sUser ) ) {
    &logTrace ( 20010,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20011,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # Remove preexisting IPC resources..
  #-----------------------------------
  if($sComponent eq "DSEngine"){
     if(RemoveIPCs($sComponent, $sInfoSphereInstance)){
        &logTrace ( 20152,
        "<%s> Confirmed IPC resources for InfoSphere instance [%s] do not exist",
        $sMyName, $sInfoSphereInstance);
     }
     else {
        &logError ( 20153,
        "<%s> Could not confirm removal of IPC resources for InfoSphere instance [%s]",
        $sMyName, $sInfoSphereInstance);
     }
  }

  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 );
  
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  my %hCommands =  ( DSEngine        => "$sInfoSphereInstanceDir/Server/DSEngine/bin/uv", 
                     ASBNodeAgent    => "$sInfoSphereInstanceDir/ASBNode/bin/NodeAgents.sh",
                     ASBLoggingAgent => "$sInfoSphereInstanceDir/ASBNode/bin/NodeAgents.sh"
                   );

  my %hCommandArgs = ( DSEngine        => "-admin -start",
                       ASBNodeAgent    => "start",
                       ASBLoggingAgent => "startLogging"
                     );

  $sCommand     = $hCommands{$sComponent};
  $sCommandArgs = $hCommandArgs{$sComponent};


  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20012,
  "<%s> Starting InfoSphere instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20013,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif( $::gbUseSystemD && SystemD::DetectSystemd() ) {
    my $WhichCommand = "Start" . $sComponent;
    if( $iReturnCode = SystemD::RunSystemdWithEnv($sUser, $sCommand, $sCommandArgs, $sEnvFile, $iTimeout, $WhichCommand, undef) ){
      &logTrace ( 20169,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  elsif ( $iReturnCode =
    InfoSphere::ExecLocalCmd ( $sUser, 
                               $sEnvFile, 
                               $sCommand, 
                               $sCommandArgs, 
                               \$iExitCode, 
                               $iTimeout, 
                               \@lsCmdout, 
                               undef ) ) {
    &logTrace ( 20014,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20015,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         InfoSphere::GetProcState ( \@liPids,
                                    $sComponent,
                                    $sHostname,
                                    $sInfoSphereInstance,
                                    $sInfoSphereInstanceDir,
                                    $iPort,
                                    $sUser,
                                    1 ) && (           # Running FFLM..
         $iSecondLevelMonitor  ?                              # If enabled..
         VCSDef_MonitorExit::ONLINE ==                        # Online after..
         InfoSphere::SecondLevelMonitor ( $sComponent,   # Running SLM..
                                          $sEnvFile,
                                          $sHostname,
                                          $sInfoSphereInstanceDir,
                                          $iPort,
                                          $sUser,
                                          $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20016,
      "<%s> Confirmed start-up of this instance of InfoSphere server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20017,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}

=pod
=cut
sub RemoveIPCs{

  my ($sComponent,$sInfoSphereInstance) = @_;
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20154, "Subroutine <%s> called with:
  Component                [%s]
  InfoSphereInstance       [%s]", $sMyName, $sComponent, $sInfoSphereInstance );

  my $sCommand = "/usr/bin/ipcs";
  my $sCommandArgs = "";

  my (@ipcArray, @shmArray, @semArray, @msqArray) = ();

  my $shm_flag = 0;
  my $msq_flag = 0;
  my $sem_flag = 0;

  my $sPattern = "0x".$sInfoSphereInstance;

  my ($iExitCode,$iTimeout,@lsCmdout,$iReturnCode,$pat,$iTillWhen) = ();

  my $sIdCount = 0;
  my @list = ();
 
  $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 );
  
    if ( ( $iTimeout = $iTillWhen - time ) <= 20 ) { # Get time bearings..
    &logError ( 20166,
    "<%s> Insufficient time to remove IPC resources. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
    } elsif ( $iReturnCode = InfoSphere::ExecLocalCmd ( "root","",
                                     $sCommand,
                                     $sCommandArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdout,
                                     undef ) ) {

       &logTrace ( 20155,
       "<%s> Cmd [%s] with args [%s] executed successfully",
       $sMyName, $sCommand, $sCommandArgs );

       foreach(@lsCmdout){

         	if($_ =~ /shmid/){
	           $shm_flag = 1;
        	   $sem_flag = 0;
	           $msq_flag = 0;
        	}
	        if($_ =~ /msqid/){
        	   $msq_flag = 1;
	           $shm_flag = 0;
        	   $sem_flag = 0;
	        }
	        if($_ =~ /semid/){
        	   $sem_flag = 1;
	           $msq_flag = 0;
	           $shm_flag = 0;
        	}

	        if ($_ =~ /$sInfoSphereInstance/){

        	    @ipcArray = split('\s+',$_);
	            $sIdCount = @ipcArray;

        	    if($sIdCount > 1){
                	if($sem_flag == 1){
	                   push(@semArray, $ipcArray[1]);
        	        }
                	if($msq_flag == 1){
	                   push(@shmArray, $ipcArray[1]);
        	        }
                	if($shm_flag == 1){
	                   push(@msqArray, $ipcArray[1]);
        	        }
	             }
         	}
       } # end of foreach 
  } # end of execLocalCmd

  if(scalar(@semArray)){
     if ( Proc::RemoveIPCResourcesForUser ( "root",O_IPCSEM,\@semArray ) ) {
        &logTrace (20156, "<%s> Confirmed SEM Ids [%s] do not exist on system",$sMyName, join( ' ', @semArray ));
     }
     else {
        &logError (20157, "<%s> Failed to remove all Semaphores", $sMyName ); # Verify..
        $iReturnCode = 0; # Failure..
     }
  } 

  if(scalar(@shmArray)){
     if ( Proc::RemoveIPCResourcesForUser ( "root",O_IPCSHM,\@shmArray ) ) {
        &logTrace (20158, "<%s> Confirmed SHM Ids [%s] do not exist on system",$sMyName, join( ' ', @shmArray ));
     }
     else {
        &logError (20159, "<%s> Failed to remove all Shared Memories", $sMyName ); # Verify..
        $iReturnCode = 0; # Failure..
     }
  }

  if(scalar(@msqArray)){
     if ( Proc::RemoveIPCResourcesForUser ( "root",O_IPCMSQ,\@msqArray ) ) {
        &logTrace (20160, "<%s> Confirmed MSQ Ids [%s] do not exist on system",$sMyName, join( ' ', @msqArray ));
     }
     else {
        &logError (20161, "<%s> Failed to remove all Message Queues", $sMyName ); # Verify..
        $iReturnCode = 0; # Failure..
     }
  }
   
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String Component,
                        IN String EnvFile,
                        IN String Hostname,
                        IN String InfoSphereInstance,
                        IN String InfoSphereInstanceDir,
                        IN Integer Port,
                        IN String User );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sComponent                The engine tier component that we want to start.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sEnvFile,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20018, "Subroutine <%s> called with:
  Component                [%s]
  EnvFile                  [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]",

  $sMyName,       $sComponent,        $sEnvFile,        $sHostname,
  $sInfoSphereInstance, $sInfoSphereInstanceDir, $iPort, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..
  

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  #------------------------------------------
  # Define stop command and args variables..
  #------------------------------------------

  my %hCommands = ( DSEngine        => "$sInfoSphereInstanceDir/Server/DSEngine/bin/uv",
                    ASBNodeAgent    => "$sInfoSphereInstanceDir/ASBNode/bin/NodeAgents.sh",
                    ASBLoggingAgent => "$sInfoSphereInstanceDir/ASBNode/bin/NodeAgents.sh"
                  );

  my %hCommandArgs = ( DSEngine        => "-admin -stop",
                       ASBNodeAgent    => "stopAgent",
                       ASBLoggingAgent => "stop"
                     );

  $sCommand     = $hCommands{$sComponent};
  $sCommandArgs = $hCommandArgs{$sComponent};

  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20019,
  "<%s> Stopping InfoSphere with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20020,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          InfoSphere::ExecLocalCmd ( $sUser, 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout, 
                                     undef ) ) {
    &logTrace ( 20021,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }


  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20022,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      InfoSphere::GetProcState ( \@liPids,
                             $sComponent,
                                 $sHostname,
                                 $sInfoSphereInstance,
                                 $sInfoSphereInstanceDir,
                                 $iPort,
                                 $sUser,
                                 0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20023,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20024,
        "<%s> Confirmed shutdown of InfoSphere component [%s]",
        $sMyName,$sComponent);
        last; # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting InfoSphere processes..
  #-----------------------------------------
  &logTrace ( 20025,
  "<%s> Killing InfoSphere processes for component [%s]", $sMyName, $sComponent);

  if ( InfoSphere::KillInfoSphereProcesses ( $sComponent,
                                             $sHostname,
                                             $sInfoSphereInstance,
                                             $sInfoSphereInstanceDir,
                                             $iPort,
                                             $sUser ) ) {
    &logTrace ( 20026,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20027,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
 
  # Code for removal of IPC resources for DSEngine 
  if($sComponent eq "DSEngine"){
     if(RemoveIPCs($sComponent, $sInfoSphereInstance)){
        &logTrace ( 20162,
        "<%s> Confirmed IPC resources for InfoSphere instance [%s] do not exist",
        $sMyName, $sInfoSphereInstance);
     }
     else {
        &logError ( 20163,
        "<%s> Could not confirm removal of IPC resources for InfoSphere instance [%s]",
        $sMyName, $sInfoSphereInstance);
     }
  }   
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String Component,
                        IN String EnvFile,
                        IN String Hostname,
                        IN String InfoSphereInstance,
                        IN String InfoSphereInstanceDir,
                        IN Integer Port,
                        IN String User );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sComponent                The engine tier component that we want to start.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sComponent,
       $sEnvFile,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20028, "Subroutine <%s> called with:
  CleanReason              [%s]
  Component                [%s]
  EnvFile                  [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]",

  $sMyName,  $sCleanReason,  $sComponent,  $sEnvFile,   $sHostname,
  $sInfoSphereInstance, $sInfoSphereInstanceDir, $iPort, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  
  my %hCommands = ( DSEngine        => "$sInfoSphereInstanceDir/Server/DSEngine/bin/uv",
                    ASBNodeAgent    => "$sInfoSphereInstanceDir/ASBNode/bin/NodeAgents.sh",
                    ASBLoggingAgent => "$sInfoSphereInstanceDir/ASBNode/bin/NodeAgents.sh"
                  );

  my %hCommandArgs = ( DSEngine        => "-admin -stop",
                       ASBNodeAgent    => "stopAgent",
                       ASBLoggingAgent => "stop"
                     );

  $sCommand     = $hCommands{$sComponent};
  $sCommandArgs = $hCommandArgs{$sComponent};


  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20029,
    "<%s> Attempting to stop InfoSphere with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20030,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( InfoSphere::ExecLocalCmd ( $sUser, 
                                       $sEnvFile, 
                                       $sCommand, 
                                       $sCommandArgs, 
                                       \$iExitCode, 
                                       $iTimeout, 
                                       \@lsCmdout, 
                                       undef ) ) {
      &logTrace ( 20031, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #-----------------------------------------
  # Whack any persisting InfoSphere processes..
  #-----------------------------------------
  &logTrace ( 20032,
  "<%s> Killing InfoSphere processes for this instance", $sMyName );

  if ( InfoSphere::KillInfoSphereProcesses ( $sComponent,
                                             $sHostname,
                                             $sInfoSphereInstance,
                                             $sInfoSphereInstanceDir,
                                             $iPort,
                                             $sUser ) ) {
    &logTrace ( 20033,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20034,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  # Code for Removal of IPC Resources
  if($sComponent eq "DSEngine"){
     if(RemoveIPCs($sComponent, $sInfoSphereInstance)){
        &logTrace ( 20164,
        "<%s> Confirmed IPC resources for InfoSphere instance [%s] do not exist",
        $sMyName, $sInfoSphereInstance);
     }
     else {
        &logError ( 20165,
        "<%s> Could not confirm removal of IPC resources for InfoSphere instance [%s]",
        $sMyName, $sInfoSphereInstance);
     }
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String Component,
                                   IN String Hostname,
                                   IN String InfoSphereInstance,
                                   IN String InfoSphereInstanceDir,
                                   IN Integer Port,
                                   IN String User );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sComponent                The engine tier component that we want to start.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $sPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $sPort,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20035, "Subroutine <%s> called:
  Component                [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]",

  $sMyName,    $sComponent,    $sHostname,    $sInfoSphereInstance,
  $sInfoSphereInstanceDir, $sPort, $sUser );
  
  my @liDummyPids = ();
  my $iReturnCode = InfoSphere::GetProcState ( \@liDummyPids,
                                               $sComponent,
                                               $sHostname,
                                               $sInfoSphereInstance,
                                               $sInfoSphereInstanceDir,
                                               $sPort,
                                               $sUser,
                                               1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20037,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20038,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20039,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String Component,
                               IN String Hostname,
                               IN String InfoSphereInstance,
                               IN String InfoSphereInstanceDir,
                               IN Integer Port,
                               IN String User,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sComponent                The engine tier component that we want to start.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sComponent,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20040, "Subroutine <%s> called with:
  OnlinePids               [%s]
  Component                [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]
  UseFFLM                  [%s]",

  $sMyName,      $rliOnlinePids,      $sComponent,      $sHostname,
  $sInfoSphereInstance,  $sInfoSphereInstanceDir,  $iPort,  $sUser,
  $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues, %hiPids,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  # Below check is added because for engine tier component "DSEngine", the 
  # dsrpcd process is started by logging in as "dsadm", but the process is run as root
  # as its setuid 's' bit is set.

  if ( $sComponent eq "DSEngine" ){
       $sUser = "root";
  }

  if ( ( 0 == ( $iUid = getpwnam $sUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20041, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( InfoSphere::GetSrvrDetails ( $sComponent, 
                                      $sHostname, 
                                      $sInfoSphereInstance, 
                                      $sInfoSphereInstanceDir, 
                                      $iPort, 
                                      $sUser, 
                                      \%hsSrvrDetails ) ) {
      &logTrace ( 20042,
      "<%s> Got details for monitoring InfoSphere instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20043,
      "<%s> Failed to get details for monitoring InfoSphere instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nInfoSphere process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = InfoSphere::FFLMStatus ( $sPidFile,
                                         $sProc,
                                         $iUid,
                                         \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20044,
      "<%s> InfoSphere instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20045,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20046,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20047,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20048,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20049, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                  unshift ( @$rliOnlinePids, $iPid );         # Store..
    
                  &logTrace ( 20050,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  $hiProcPids { $sKey } = $iPid;        # Store..
                  
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20051,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20052,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( InfoSphere::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20053,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20054,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20055,
    "<%s> Found [%s] PIDs [%s] for this instance of InfoSphere.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20056,
    "<%s> InfoSphere process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String Component,
                          IN String EnvFile,
                          IN String Hostname,
                          IN String InfoSphereInstanceDir,
                          IN String MonitorProgram,
                          IN Integer Port,
                          IN Integer SecondLevelMonitor,
                          IN String User,
                          IN Integer $level_two );

 Description: Performs the in-depth monitoring

 Parameters :

 $sComponent                The engine tier component that we want to start.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $sMonitorProgram           Full path to external monitor program.
 $iPort                     Port allocated for the specific engine tier component. 
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 $level_two                Level Two Monitoring when IMF Enabled
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sEnvFile,
       $sHostname,
       $sInfoSphereInstanceDir,
       $sMonitorProgram,
       $iPort,
       $iSecondLevelMonitor,
       $sUser,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20057, "Subroutine <%s> called with:
  Component                [%s]
  EnvFile                  [%s]
  Hostname                 [%s]
  InfoSphereInstanceDir    [%s]
  MonitorProgram           [%s]
  Port                     [%s]
  SecondLevelMonitor       [%s]
  User                     [%s]
  LevelTwoMonitor          [%s]",

  $sMyName,       $sComponent,        $sEnvFile,        $sHostname,
  $sInfoSphereInstanceDir,        $sMonitorProgram,         $iPort,
  $iSecondLevelMonitor, $sUser, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( -300 == $level_two ) {
     $level_two = VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor );  
     &logTrace ( 20058, "<%s> Traditional second level monitoring, SLM flag for this monitor set to [%s] ", $sMyName, $level_two  );
  }

  if ( $level_two >= 1 ) 
  {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         InfoSphere::DelayedAttributeCheck ( $sComponent,
                                             $sEnvFile,
                                             $sHostname,
                                             $sInfoSphereInstanceDir,
                                             $iPort,
                                             $sUser ) ) ) {
      &logTrace ( 20059,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20060,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20061,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         InfoSphere::SecondLevelMonitor ( $sComponent,
                                          $sEnvFile,
                                          $sHostname,
                                          $sInfoSphereInstanceDir,
                                          $iPort,
                                          $sUser,
                                          $iTillWhen ) ) ) {

      &logTrace ( 20062,
      "<%s> InfoSphere instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20063,
      "<%s> InfoSphere instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + VCSagentFW::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         InfoSphere::ExternalMonitor ( $sUser,
                                       $sEnvFile,
                                       $sExtCmd,
                                       $sExtCmdArgs,
                                       $iTillWhen ) ) ) {
      &logTrace ( 20064,
      "<%s> InfoSphere instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20065,
      "<%s> InfoSphere instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String Component,
                                        IN String EnvFile,
                                        IN String Hostname,
                                        IN String InfoSphereInstanceDir,
                                        IN Integer Port,
                                        IN String User );

 Description : Just In Time Validation

 Parameters :

 $sComponent                The engine tier component that we want to start.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sEnvFile,
       $sHostname,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20066, "Subroutine <%s> called with:
  Component                [%s]
  EnvFile                  [%s]
  Hostname                 [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]",

  $sMyName,       $sComponent,        $sEnvFile,        $sHostname,
  $sInfoSphereInstanceDir, $iPort, $sUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( Component EnvFile Hostname InfoSphereInstanceDir Port User );
  my @lsReqdArgs = qw ( Component Hostname InfoSphereInstanceDir Port User );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20067,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #--------------------------------------------
  # Validate parameter InfoSphereInstanceDir..
  #--------------------------------------------
  if ( ! -d $sInfoSphereInstanceDir ) {
    &logError ( 20068,
    "<%s> InfoSphereInstanceDir [%s] directory does not exist",
    $sMyName, $sInfoSphereInstanceDir );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  # Only the DSEngine requires the EnvFile to startup.
  if ( $sComponent eq "DSEngine" ){
       if ( ! -f $sEnvFile ) {
            &logError ( 20069,"<%s> EnvFile [%s] is not a plain file",$sMyName, $sEnvFile );
            $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
       }  
  }
  #---------------------------
  # Validate parameter User..
  #---------------------------
  if ( ! defined getpwnam $sUser ) {
    &logError ( 20070,
    "<%s> User [%s] is not a valid UNIX login",
    $sMyName, $sUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------------------
  # Check if virtual hostname is up locally..
  #------------------------------------------
  if($sHostname){
     unless ( Comms::IsIPLocal ($sHostname) ) {
       &logError ( 20036,
       "<%s> Virtual Host [%s] not up locally", $sMyName, $sHostname );
       return VCSDef_MonitorExit::UNKNOWN;
     }
  }

  if($iPort == 0){
       &logError ( 20167,
       "<%s> Port cannot be set to a 0 for component [%s]", $sMyName, $sComponent );
       return VCSDef_MonitorExit::UNKNOWN;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String Component,
                                     IN String EnvFile,
                                     IN String Hostname,
                                     IN String InfoSphereInstanceDir,
                                     IN Integer Port,
                                     IN String User,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sComponent                The engine tier component that we want to start.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sEnvFile,
       $sHostname,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20071, "Subroutine <%s> called with:
  Component                [%s]
  EnvFile                  [%s]
  Hostname                 [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]
  TillWhen                 [%s]",

  $sMyName,       $sComponent,        $sEnvFile,        $sHostname,
  $sInfoSphereInstanceDir, $iPort, $sUser, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;
  
  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #------------------------------------
  # Check if port can be connected to..
  #------------------------------------
  if ( Comms::CheckPort ( $sHostname,
                          $iPort ) ) {
    &logTrace ( 20072,
    "<%s> Engine tier component [%s] listening on Host [%s] Port [%s]: Socket method",
    $sMyName, $sComponent, $sHostname, $iPort );
  }
  else {
    &logError ( 20073,
    "<%s> Engine tier component [%s] NOT listening on Host [%s] Port [%s]: Socket method",
    $sMyName, $sComponent, $sHostname, $iPort );

    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Return, unless online..
  #------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String User,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 $sEnvFile                  Absolute path to the DataStage environment file that will be sourced for starting the DSEngine.
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20076, "Subroutine <%s> called with:
  User                     [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20077,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20078,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20079,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20080,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20081,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  InfoSphere::ExecLocalCmd ( $sUser, 
                             $sEnvFile, 
                             $sCmd, 
                             $sCmdArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20082,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20083,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20084, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20085, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20086,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20087,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20088,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20089, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20090, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20091, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20092, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20093,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20094,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20095,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( InfoSphere::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20096,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20097,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20098,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20099,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20100, "<%s> PID [%s] from PID file [%s] does not exist",
    $sMyName, $iPid, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillInfoSphereProcesses ( IN String Component,
                                           IN String Hostname,
                                           IN String InfoSphereInstance,
                                           IN String InfoSphereInstanceDir,
                                           IN Integer Port,
                                           IN String User );

 Description  : Kills processes for this instance of InfoSphere
 
 Parameters :

 $sComponent              The engine tier component that we want to start.
 $sHostname               Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance     The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir  The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                   Port allocated for the specific engine tier component. 
 $sUser                   Specifies the OS user required to start the DataStage engine or the ASBAgent.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillInfoSphereProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20101, "Subroutine <%s> called with:
  Component                [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]",

  $sMyName,    $sComponent,    $sHostname,    $sInfoSphereInstance,
  $sInfoSphereInstanceDir, $iPort, $sUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'InfoSphere::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    InfoSphere::GetProcState ( \@liGoners,
                           $sComponent,
                               $sHostname,
                               $sInfoSphereInstance,
                               $sInfoSphereInstanceDir,
                               $iPort,
                               $sUser,
                               0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20102,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20103,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20104,
    "<%s> Could not locate any process this instance of InfoSphere",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sOutputFilename ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20105, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20106,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20107,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20108,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20109,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout);

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20110,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20111,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20112,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20113,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20114, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20115,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if ( "solaris" eq $sSwitch ) {
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }
  elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20116, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20117,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20118,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String Component,
                                  IN String Hostname,
                                  IN String InfoSphereInstance,
                                  IN String InfoSphereInstanceDir,
                                  IN Integer Port,
                                  IN String User,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sComponent                The engine tier component that we want to start.
 $sHostname                 Hostname/IPv4 address of virtual host for DataStage server. 
 $sInfoSphereInstance       The instance name of IBM InfoSphere DataStage server. 
 $sInfoSphereInstanceDir    The directory where IBM InfoSphere DataStage server instance is installed.
 $iPort                     Port allocated for the specific engine tier component. 
 $sUser                     Specifies the OS user required to start the DataStage engine or the ASBAgent.
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sComponent,
       $sHostname,
       $sInfoSphereInstance,
       $sInfoSphereInstanceDir,
       $iPort,
       $sUser,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20119, "Subroutine <%s> called with:
  Component                [%s]
  Hostname                 [%s]
  InfoSphereInstance       [%s]
  InfoSphereInstanceDir    [%s]
  Port                     [%s]
  User                     [%s]
  SrvrDetails              [%s]",

  $sMyName,    $sComponent,    $sHostname,    $sInfoSphereInstance,
  $sInfoSphereInstanceDir, $iPort, $sUser, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat, $sSwitch ) = ();

  $sSwitch = $^O;

  my @lsSrchPatterns = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  if($sComponent eq "DSEngine"){
     if($sInfoSphereInstance eq "ade"){
        @lsSrchPatterns = ( qr@$sInfoSphereInstanceDir/Server/DSEngine/bin/dsrpcd@ );
     }else{
        @lsSrchPatterns = ( qr!$sInfoSphereInstanceDir/Server/DSEngine/bin/$sInfoSphereInstance.*dsrpcd! );
     }
  } elsif($sComponent eq "ASBNodeAgent") {
     if($sSwitch eq "solaris"){
        @lsSrchPatterns = ( qr!$sInfoSphereInstanceDir/ASBNode/apps/jre/bin/sparcv9/java.*asb.agent*! );
     }else{
        @lsSrchPatterns = ( qr!$sInfoSphereInstanceDir/((ASBNode/apps)|jdk)/jre/bin/java.*asb.?agent*! );
     }
  } elsif($sComponent eq "ASBLoggingAgent") {
     @lsSrchPatterns = ( qr!$sInfoSphereInstanceDir/ASBNode/apps/jre/bin/java.*logging.agent*! );
  }

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  my $sOldMask = umask ( 000 );

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20120, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
    chmod (0777, "/tmp/.VRTS$sResourceType");
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0777 ) ) {
    &logWarn ( 20121, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  umask ( $sOldMask );
  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
